from contextlib import asynccontextmanager
from fastapi import FastAPI, APIRouter
from user_auth_zxw.apis import api_用户权限_增加
from user_auth_zxw.apis import api_JWT, api_用户权限_验证
from user_auth_zxw.apis.api_登录注册 import 微信H5快捷登录, 账号密码注册登录, 手机注册登录
import logging
from fastapi.responses import JSONResponse
from user_auth_zxw.db import Base, engine


@asynccontextmanager
async def lifespan(app: FastAPI):
    # 启动: 在 FastAPI 应用启动时创建表结构
    async with engine.begin() as conn:
        await conn.run_sync(Base.metadata.create_all)
    print("All tables created if not existing.")

    # 使用Yield，控制程序回到FastAPI服务
    yield

    # 关闭逻辑: close connections, etc.
    await engine.dispose()


router = APIRouter(lifespan=lifespan)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# @router.exception_handler(Exception)
# async def global_exception_handler(request, exc):
#     logger.error(f"Unhandled exception: {exc}")
#     return JSONResponse(status_code=500, content={"detail": "Internal server error"})


# Register API routers
router.include_router(api_用户权限_增加.router, tags=["用户权限管理"])
router.include_router(api_用户权限_验证.router, tags=["用户权限管理"])
router.include_router(api_JWT.router, tags=["JWT"])
router.include_router(手机注册登录.api_注册登录.router, tags=["注册登录-手机邮箱"])
router.include_router(微信H5快捷登录.api_注册登录.router, tags=["登录注册-微信公众号"])
router.include_router(账号密码注册登录.api_注册登录.router, tags=["登录注册-账号密码"])

# Run the application
if __name__ == "__main__":
    import uvicorn
    from fastapi import FastAPI

    app = FastAPI()
    app.include_router(router)

    uvicorn.run(router, host="localhost", port=8101)



"""
# File       : config.py
# Time       ：2024/8/20 下午5:25
# Author     ：xuewei zhang
# Email      ：shuiheyangguang@gmail.com
# version    ：python 3.12
# Description：
"""
from pathlib import Path

if Path("configs/config_user_auth.py").exists():
    from configs.config_user_auth import *
else:
    import os
    from datetime import timedelta

    # postgresql数据库
    DATABASE_URL = os.environ.get('USER_CENTER_DATABASE_URL')  # 读取docker配置的环境变量
    if not DATABASE_URL:
        DATABASE_URL = "postgresql+asyncpg://...:...@localhost/..."
        os.environ['USER_CENTER_DATABASE_URL'] = DATABASE_URL  # 设置环境变量, 以便在docker、alembic中读取

    # redis数据库 redis://[[username]:[password]]@localhost:6379/0
    redisUrl = "redis://[[username]:[password]]@localhost:6379/0"


    # jwt setting
    class JWT:
        SECRET_KEY = "...your_secret_key..."
        ALGORITHM = "HS256"
        expire_time = timedelta(days=7)


    # 微信公众号
    class WeChatPub:
        app_id = "...your_app_id..."
        app_secret = "..."
        scope = "snsapi_login"
        scope_qrcode_login = "snsapi_login"  # 二维码登录必须用snsapi_login
        state = "your_custom_state"  # 用于防止CSRF
        接口配置信息_Token = "..."  # 自己去微信公众号设置


    # 微信小程序
    class WeChatMini:
        app_id = "..."
        app_secret = "..."


    # 阿里云
    class Aliyun:
        ali_access_key_id = "..."
        ali_access_key_secret = "..."
        ali_secretNo_pool_key = "..."


    class AliyunSMS:
        ali_access_key_id = "..."
        ali_access_key_secret = "..."
        ali_secretNo_pool_key = "..."


    # 支付专用
    class WeChatPay:
        MCH_ID = '...'
        SECRET = '...'
        NONCE_STR = '...'
        KEY = '...'
        PAYMENT_NOTIFY_URL = 'https://tongsheng.natapp4.cc/wxpay_recall'
        REFUND_NOTIFY_URL = 'https://tongsheng.natapp4.cc/wxpay_recall'


    # 发送邮件
    class Email:
        sender = '...@163.com'  # 发件人
        server = 'smtp.163.com'  # 所使用的用来发送邮件的SMTP服务器
        username = '...@163.com'  # 发送邮箱的用户名和授权码（不是登录邮箱的密码）
        password = '...'  # 服务器: MVQDSPUQATBDOIFU / 自用电脑: FFHBMPJSXXFEEZIK


    # AES密码密匙
    class AESKey:
        key_web = "..."
        key_local = "..."



"""
# File       : models.py
# Time       ：2024/8/20 下午5:20
# Author     ：xuewei zhang
# Email      ：shuiheyangguang@gmail.com
# version    ：python 3.12
# Description：
"""
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy import Column, Integer, String, ForeignKey, Table, UniqueConstraint
from passlib.context import CryptContext
from user_auth_zxw.apis.schemas import Payload

Base = declarative_base()  # 创建一个基类,

user_role_table = Table(
    'user_role',
    Base.metadata,
    Column('user_id', Integer, ForeignKey('users.id')),
    Column('role_id', Integer, ForeignKey('roles.id'))
)

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


class User(Base):
    __tablename__ = "users"
    id = Column(Integer, primary_key=True, index=True)
    # 账号 - 必填
    username = Column(String, unique=True, index=True, nullable=True)
    email = Column(String, unique=False, nullable=True)
    phone = Column(String, unique=True, index=True, nullable=True)
    openid = Column(String, unique=True, index=True, nullable=True)
    # 密码，可以为空，因为有的用户可能是第三方登录
    hashed_password = Column(String, nullable=True)
    #
    nickname = Column(String, nullable=True)
    notes = Column(String, nullable=True)
    # secondary作用是指定中间表，back_populates作用是指定反向关系
    # 中间表用于存储多对多关系，反向关系用于在查询时，通过一个表查询另一个表
    # lazy='selectin'表示在查询时，会将关联的表一次性查询出来，而不是按需查询. 可以在fastapi中避免由于session关闭导致的查询异常。
    roles = relationship("Role", secondary=user_role_table, back_populates="users", lazy='selectin')

    def verify_password(self, password: str) -> bool:
        # 密码为空代表是第三方登录
        if password == "":
            return False
        return pwd_context.verify(password, self.hashed_password)

    @staticmethod
    def hash_password(password: str) -> str:
        return pwd_context.hash(password)

    def to_payload(self) -> Payload:
        return Payload(sub=self.username,
                       username=self.username,
                       nickname=self.nickname,
                       roles=[
                           {"role_name": role.name,
                            "app_id": role.app_id,
                            "app_name": role.app.name} for role in self.roles]
                       )


class Role(Base):
    __tablename__ = 'roles'
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, unique=True)
    app_id = Column(Integer, ForeignKey('apps.id'))
    app = relationship("App", back_populates="roles", lazy='selectin')
    users = relationship("User", secondary=user_role_table, back_populates="roles")


class App(Base):
    __tablename__ = 'apps'
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, unique=True)
    roles = relationship("Role", back_populates="app")
    UniqueConstraint('name', name='uq_app_name')



"""
# File       : __init__.py.py
# Time       ：2024/8/22 16:31
# Author     ：xuewei zhang
# Email      ：shuiheyangguang@gmail.com
# version    ：python 3.12
# Description：
"""
from .db_model.model import RefreshToken
from .jwt import create_jwt_token, get_current_user
from .jwt_刷新管理 import create_refresh_token



"""
# File       : __init__.py
# Time       ：2024/8/23 05:54
# Author     ：xuewei zhang
# Email      ：shuiheyangguang@gmail.com
# version    ：python 3.12
# Description：
"""
from .models import Base
from .get_db import get_db, engine, sync_engine



from setuptools import setup, find_packages

setup(
    name="user_auth_zxw",
    version="0.0.8",
    packages=find_packages(),
    include_package_data=True,  # 确保包含包内的非Python文件
    package_data={
        '': ['**/*.vue', '**/*.ts'],  # 包含所有子文件夹中的.vue和.ts文件
    },
    install_requires=[
        'pycryptodome>=3.20.0,<=3.22.0',
        'fastapi>=0.112.0,<0.113',
        'jose>=1.0.0,<1.1.0',
        'aiohttp>=3.10.5,<3.11.0',
        'httpx>=0.23.3,<=0.27.0',
        'sqlalchemy==2.0.32',
        'greenlet==3.0.3',
        'databases==0.9.0',
        'python-jose==3.3.0',
        'passlib==1.7.4',
        'asyncpg==0.29.0',
        'uvicorn>=0.30.0,<0.31.0',
        'python-multipart==0.0.9',
        'bcrypt==4.2.0',
        'app-tools-zxw>=1.0.71',
        'alibabacloud_dysmsapi20170525==3.0.0'
    ],
    author="薛伟的小工具",
    author_email="",
    description="用户权限验证工具包",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/sunshineinwater/",
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
# File       : 角色.py
# Time       ：2024/8/20 下午6:09
# Author     ：xuewei zhang
# Email      ：shuiheyangguang@gmail.com
# version    ：python 3.12
# Description：
"""
from fastapi import APIRouter, Depends
from sqlalchemy.future import select
from sqlalchemy.ext.asyncio import AsyncSession

from user_auth_zxw.db.models import User, App
from user_auth_zxw.db.get_db import get_db
from user_auth_zxw.SDK_jwt.jwt import get_current_user
from user_auth_zxw.apis.schemas import 请求_验证角色_from_header, 返回_验证角色_from_header

from app_tools_zxw.Errors.api_errors import HTTPException_AppToolsSZXW, ErrorCode

router = APIRouter(prefix="/roles")


def require_role(role_name: str, app_name: str):
    """
    本函数主要作用：检查用户是否有某个角色，如果没有则返回403，如果有则返回用户信息
    使用方法：在需要检查角色的路由上加上Depends(require_role("admin", "app1"))
    使用示例：下方的get_admin_data函数
    主要流程：1. 通过jwt验证获取用户信息
            2. 通过app_name获取应用信息
            3. 判断用户是否有该角色
    :param role_name:
    :param app_name:
    :return:
    """

    async def role_checker(user: User = Depends(get_current_user), db: AsyncSession = Depends(get_db)):
        result_app = await db.execute(select(App).filter(App.name == app_name))
        app = result_app.scalar_one_or_none()
        if app is None:
            raise HTTPException_AppToolsSZXW(
                error_code=ErrorCode.应用未找到,
                detail="应用未找到",
                http_status_code=403
            )

        if not any(role.name == role_name and role.app_id == app.id for role in user.roles):
            raise HTTPException_AppToolsSZXW(
                error_code=ErrorCode.权限不足,
                detail="权限不足",
                http_status_code=403
            )
        return user

    return role_checker


def require_roles(role_names: list[str], app_name: str):
    """
    本函数主要作用：检查用户是否有某个角色，如果没有则返回403，如果有则返回用户信息
    使用方法：在需要检查角色的路由上加上Depends(require_role("admin", "app1"))
    使用示例：下方的get_admin_data函数
    主要流程：1. 通过jwt验证获取用户信息
            2. 通过app_name获取应用信息
            3. 判断用户是否有该角色
    :param role_names:
    :param app_name:
    :return:
    """

    async def role_checker(
            user: User = Depends(get_current_user),
            db: AsyncSession = Depends(get_db)
    ):
        result_app = await db.execute(select(App).filter(App.name == app_name))
        app = result_app.scalar_one_or_none()
        if app is None:
            raise HTTPException_AppToolsSZXW(
                error_code=ErrorCode.应用未找到,
                detail="应用未找到",
                http_status_code=403
            )

        if not any(role.name in role_names and role.app_id == app.id for role in user.roles):
            raise HTTPException_AppToolsSZXW(
                error_code=ErrorCode.权限不足,
                detail="权限不足",
                http_status_code=403
            )
        return user

    return role_checker


@router.post("/role-auth/", response_model=返回_验证角色_from_header)
async def 验证角色_from_header(request: 请求_验证角色_from_header,
                               user: User = Depends(get_current_user),
                               db: AsyncSession = Depends(get_db)):
    """
    权限验证函数
    :param request: role_name,app_name
    :param user:
    :param db:
    :return:
    """
    print(user.to_payload())
    user: User = await require_role(request.role_name, request.app_name)(user, db)
    return {'status': True}


@router.get("/admin-data/")
async def demo_验证admin角色(user: User = Depends(require_role("admin", "app1"))):
    """
    权限验证函数的使用示例
    :param user:
    :return:
    """
    return {"data": "This is admin data"}



"""
# File       : jwt.py
# Time       ：2024/8/20 下午5:27
# Author     ：xuewei zhang
# Email      ：shuiheyangguang@gmail.com
# version    ：python 3.12
# Description：
    1. create_jwt_token() 生成JWT令牌
    2. get_current_user() 从当前请求中，读取令牌，并验证用户身份。
"""
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError, jwt
from datetime import datetime, timedelta
from sqlalchemy.future import select
from sqlalchemy.ext.asyncio import AsyncSession
# 用户自定义部分
from user_auth_zxw.config import JWT
from user_auth_zxw.db.get_db import get_db
from user_auth_zxw.db.models import User

ACCESS_TOKEN_EXPIRE_MINUTES = 24 * 60 * 7  # 默认令牌过期时间，单位是分钟
# tokenUrl是登录页面的URL, 用于文档测试
oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/api/account/login-form/")


def create_jwt_token(payload: dict, expires_delta: timedelta = None):
    """
    任意使用
    生成JWT令牌
    :param payload: 通常包含以下信息：
                    sub (subject)：用户的唯一标识符
                    name：用户的名字
                    iat (issued at)：JWT的签发时间
                    exp (expiration)：JWT的过期时间
    :param expires_delta: 过期时间。如果不设置，将使用默认的过期时间：ACCESS_TOKEN_EXPIRE_MINUTES
    :return:
    """
    to_encode = payload.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(to_encode, JWT.SECRET_KEY, algorithm=JWT.ALGORITHM)
    return encoded_jwt


def check_jwt_token(token: str):
    """
    任意使用
    验证JWT令牌
    :param token:
    :return:
    """
    try:
        payload = jwt.decode(token, JWT.SECRET_KEY, algorithms=[JWT.ALGORITHM])
        return payload
    except JWTError:
        return None


async def get_current_user(token: str = Depends(oauth2_scheme),
                           db: AsyncSession = Depends(get_db)) -> User:
    """
    仅微服务可用，依赖数据库。
    从当前请求中，读取令牌，并验证用户身份。

    使用方法：在需要验证用户身份的路由上加上Depends(get_current_user)
            如： async def role_checker(user: User = Depends(get_current_user)):
                    ...

    前端请求格式：
        请求头的格式应该是Authorization，前端发送请求时，应该在请求头中添加：Authorization: Bearer token。

        uniApp中的添加方式如下：
    ```
        uni.request({
            url: 'https://...'
            method: 'POST',
            header: {
                'Authorization': 'Bearer ' + token
            },
            data: {
                ...
            },
            success: (res) => {
                ...
            }
        })
    ```
    :param token:
    :param db:
    :return:
    """
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"Authenticate": "Bearer "},
    )
    # print(f"token: {token}")
    try:
        payload = jwt.decode(token, JWT.SECRET_KEY, algorithms=[JWT.ALGORITHM])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_exp = payload.get("exp")
        if datetime.utcfromtimestamp(token_exp) < datetime.utcnow():
            raise credentials_exception

        result = await db.execute(select(User).filter(User.username == username))
        user: User = result.scalar_one_or_none()
        if user is None:
            raise credentials_exception
    except JWTError as e:
        print(f"jwt解码错误,JWTError: {e}")
        raise JWTError(e)
    except Exception as e:
        print(f"jwt解码错误,Exception: {e}")
        raise Exception(e)
    return user
